(define-module (sx-ext divide))
(use-modules  (sx entry) (sx selection-basic) (sx selection)
              (sx property) (sx basic) (sx undo)
              (sx gui-gl) (sx gui-entry) (sx gui-optmenu)
)


(export command-divide)


; add the command to the sx system
(entry-add-command-list '("divide"))



(define (command-divide)
  ; setup selection mechanism
  (selection-setup selection-object
                   operation-divide
                   "Select two points:" 
                   *OPTION_SELECT*
                   'select_obj
                   ; user must select points
                   '(no_empty type_point)
                   #f)
)



(define (operation-divide)
  (case *SELECT_STATE*
    ( (objects)
      (if (= (length *SELECT_LIST*) 2)
        (let* ( (objects-to-divide *SELECT_LIST*) )
          ; change selection mechanism
          (selection-setup selection-integer operation-divide
                           "Number of divisions:"
                           *OPTION_EMPTY* 
                           'view
                           '(no_empty no_zero no_negative) #f)
          (set! *SELECT_LIST* objects-to-divide)
        )
        (gui-entry-add-history " *Two points are required*")
      )
    )
    ( (integer)
      ; mark the beginning of the operation
      (undo-add-op-mark)
      (let*
        ( (n *SELECT_ANSWER*)
          (nb (cadar *SELECT_LIST*))
          (ne (cadadr *SELECT_LIST*))
          (cob (object-property-get 'p nb 'coords))
          (coe (object-property-get 'p ne 'coords))
          (con #f) ; the new coordinates
          (nn  #f) ; the new node
          (np  ne) ; the previous node
          (n *SELECT_ANSWER*)
          (shared '())
          (elemb (object-property-get 'p nb 'parent))
          (eleme (object-property-get 'p ne 'parent))
        )
        ; check whether they share a link element
        (for-each (lambda (x)
            (if (and (member x eleme) (eq? (car x) 'l))
              (set! shared (cons x shared))
            )
          )
          elemb
        )
        ; remove the shared link elements
        (for-each (lambda (x) (link-del-basic (cadr x) #t)) shared)

        ; generate the points between the two points
        (do ((i 1 (1+ i))) ((= i n))
          (set! con 
                (map (lambda (x y) (/ (+ (* x i) (* y (- n i))) n))
                     cob coe
                ) )
          (set! nn (point-add-basic #f con #f #t))
          ; if elements were deleted add them back
          (if (not (null? shared))
            (begin
              (link-add-basic #f (list np nn) #f #t)
              (set! np nn)
            )
          )
        )
        (if (not (null? shared))
          (link-add-basic #f (list np nb) #f #t)
        )
      )
      ; clear all selection
      (selection-clear)
      ; update the screen
      (gui-update-gl-window #t)
    )
    ( (cancel)
      ; clear all selection
      (selection-clear)
      ; update the screen
      (gui-update-gl-window #t)
    )
  )
)




