; FIXME: remove (display), replace it with (gui-entry-add-history)

(define-module (sx-ext e-lib-read))
(use-modules  (ice-9 regex)
              (sx entry) (sx undo) (sx property) (sx basic) 
              (sx selection-basic) (sx view)
              (sx gui-gl) (sx gui-entry) (sx gui-optmenu)
)

(export command-e-lib-read)


; add the commands to the sx system
(entry-add-command-list '("e-lib-read"))

; private variable to store filename between selections
(define mdf-file #f)


(define (command-e-lib-read)
  (set! mdf-file #f)
  (selection-setup selection-string
                   operation-e-lib-filename
                   "Enter filename (with path, no extension):"
                   *OPTION_EMPTY* 'view #f #f)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


; reads in a non-empty line, comments are ignored
(define (e-lib-read-line fp)
  (let* ( (buffer (make-string 80))
          (curr-buf-len (string-length buffer))
          (increment 8) 
          (i 0)
          (c #f)
          (read-more #t)
        )
    (while read-more
      (set! c (read-char fp))
      (if (not (eof-object? c))
        (begin
          ; first non-space character
          (while (and (not (eof-object? c)) (char-whitespace? c))
            (set! c (read-char fp))
          )
          (cond
            ; single line comment
            ( (equal? c #\#)
              (while (not (or (eof-object? c) (equal? c #\newline)))
                (set! c (read-char fp))
              )
            )
            ; multiple line comment
            ( (and (equal? c #\/) (equal? (peek-char fp) #\*))
              ; skip until end of it
              (while (not (or (eof-object? c) 
                              (and (equal? c #\*)
                                   (equal? (peek-char fp) #\/))))
                (set! c (read-char fp))
              )
              ; skip until end of line
              (while (not (or (eof-object? c) (equal? c #\newline)))
                (set! c (read-char fp))
              )
            )
            ( else
              (set! read-more #f)
              (while (not (or (eof-object? c) (equal? c #\newline)))
                (if (>= i curr-buf-len)
                  (begin
                    (set! buffer (string-append buffer 
                                                (make-string increment)))
                    (set! curr-buf-len (string-length buffer))
                  )
                )
                (string-set! buffer i c)
                (set! i (1+ i))
                (set! c (read-char fp))
              )
            )
          )
        )
        ; eof file
        (set! read-more #f)
      )
    ) ; end of while
     
    (if (> i 0)
      (begin
;        (display "line: ")(display (substring buffer 0 i))(newline)
        (substring buffer 0 i)
      )
      'eof
    )
  ) ; end of let 
)


; reads in an optional or mandatory keyword
; optional keywords can be ignored
(define (e-lib-read-keyword fp rx-list)
  (if (cdar rx-list)
    ; if keyword is optional 
    (call-with-current-continuation
      (lambda (found-optional)
        (do ((line (e-lib-read-line fp)
                   (e-lib-read-line fp)))
            ((equal? line 'eof))
          (call-with-current-continuation
            (lambda (end-optional)
              (for-each 
                (lambda (l-rx-type)
                  (let*
                    ( (l-rx  (car l-rx-type))
                      (l-opt (cdr l-rx-type))
                      (l-m   (regexp-exec l-rx line))
                    )
                    (if l-m
                      (found-optional l-m)
                    )
                    (if (not l-opt)
                      (end-optional #f)
                    )
                  )
                )
                rx-list
              )
            )
          )
        )
        #f
      )
    )
    ; mandatory keyword
    (call-with-current-continuation
      (lambda (found-mandatory)
        (let* ( (l-m #f) )
          (do ((line (e-lib-read-line fp)
                     (e-lib-read-line fp)))
              ((equal? line 'eof))
            (set! l-m (regexp-exec (caar rx-list) line))
            (if l-m
              (begin
                (found-mandatory l-m)
              )
            )
          )
          (gui-entry-add-history "\n*ERROR*: mandatory keyword is not found\n")
          #f
        )
      )
    )
  )
)


; reads in a series of numbers with an index
; proc determines whether it is number? or integer?
(define (e-lib-read-number-series proc i n fp)
  (call-with-current-continuation
    (lambda (return-n-series)
      (let*
        ( (l-line (e-lib-read-line fp))
          (res (catch #t 
                      (lambda ()
                        (call-with-input-string
                          (string-append "("
                                         l-line
                                         ")")
                          read)
                      )
                      (lambda args #f)))
        )
        (if (not res)
          (begin
            (display "invalid line: ")
            (display l-line)(newline)
            (return-n-series #f)
          )
        )
        (if (not (integer? (car res)))
          (begin
            (display "invalid index: ")
            (display l-line)(newline)
            (return-n-series #f)
          )
        )
        (if (not (= (1+ i) (car res)))
          (begin
            (display "index out of sequence: ")
            (display l-line)(newline)
            (return-n-series #f)
          )
        )
        (if (not (= (length (cdr res)) n))
          (begin
            (display "not enough number: ")
            (display l-line)(newline)
            (return-n-series #f)
          )
        )
        (for-each (lambda (x) 
            (if (not (proc x))
              (begin
                (display "invalid line: ")
                (display l-line)(newline)
                (return-n-series #f)
              )
            )
          )
          (cdr res)
        )
        (cdr res)
      )
    )
  )
)


(define (e-lib-read-meshpoints npoint fp)
  (call-with-current-continuation
    (lambda (return-meshpoint)
      (let ( (coord #f) (node-assoc '()) )
        (do ((i 0 (1+ i))) ((= i npoint))
          (set! coord (e-lib-read-number-series number? i 3 fp))
          (if (not coord)
            (return-meshpoint #f)
          )
          ; create point, store node id for later reference
          (set! node-assoc (cons (list (1+ i)
                                       (point-add-basic #f coord #f #t) )
                                 node-assoc))
        )
        node-assoc
      )
    )
  )
)

(define (e-lib-read-links node-assoc nlink fp type)
  (call-with-current-continuation
    (lambda (return-link)
      (let ( (nodes #f) (maped-nodes #f) (nn #f) )
        (do ((i 0 (1+ i))) ((= i nlink))
          (set! nodes (e-lib-read-number-series integer? i 2 fp))
          (if (not nodes)
            (return-link #f)
          )
          (set! maped-nodes 
                (map (lambda (x)
                       (set! nn (assoc x node-assoc))
                       (if (not nn) (return-link #f))
                       (cadr nn)
                     )
                     nodes
                )
          )
          ; ready to use the nodes
          (link-add-basic #f maped-nodes 
                          (list (list 'elemtype 'integer type)) #t)
        )
      )
      #t
    )
  )
)

(define (e-lib-read-triangs node-assoc ntriang fp type)
  (call-with-current-continuation
    (lambda (return-triang)
      (let ( (nodes #f) (maped-nodes #f) (nn #f) )
        (do ((i 0 (1+ i))) ((= i ntriang))
          (set! nodes (e-lib-read-number-series integer? i 3 fp))
          (if (not nodes)
            (return-triang #f)
          )
          (set! maped-nodes 
                (map (lambda (x)
                       (set! nn (assoc x node-assoc))
                       (if (not nn) (return-triang #f))
                       (cadr nn)
                     )
                     nodes
                )
          )
          ; ready to use the nodes
          (triang-add-basic #f maped-nodes 
                            (list (list 'elemtype 'integer type)) #t)
        )
      )
      #t
    )
  )
)

(define (e-lib-read-quads node-assoc nquad fp)
  (call-with-current-continuation
    (lambda (return-quad)
      (let ( (nodes #f) (maped-nodes #f) (nn #f) )
        (do ((i 0 (1+ i))) ((= i nquad))
          (set! nodes (e-lib-read-number-series integer? i 4 fp))
          (if (not nodes)
            (return-quad #f)
          )
          (set! maped-nodes 
                (map (lambda (x)
                       (set! nn (assoc x node-assoc))
                       (if (not nn) (return-quad #f))
                       (cadr nn)
                     )
                     nodes
                )
          )
          ; ready to use the nodes
          (quad-add-basic #f maped-nodes #f #t)
        )
      )
      #t
    )
  )
)

; reads the e-lib mesh, 
; basic geometry and topological information
(define (e-lib-read-mesh fp)
  (let* 
    ( (line #f)
      (rx-point   (make-regexp "NMESHPOINTS[ \t]+"))
      (rx-coord   (make-regexp "MESHPOINT_COORDINATES[ \t]*"))
      (rx-link1   (make-regexp "NELEMENTS_LINK1[ \t]+"))
      (rx-link2   (make-regexp "NELEMENTS_LINK2[ \t]+"))
      (rx-link3   (make-regexp "NELEMENTS_LINK3[ \t]+"))
      (rx-link4   (make-regexp "NELEMENTS_LINK4[ \t]+"))
      (rx-link5   (make-regexp "NELEMENTS_LINK5[ \t]+"))
      (rx-triang1 (make-regexp "NELEMENTS_TRIANG1[ \t]+"))
      (rx-triang3 (make-regexp "NELEMENTS_TRIANG3[ \t]+"))
      (rx-triang4 (make-regexp "NELEMENTS_TRIANG4[ \t]+"))
      (rx-triang5 (make-regexp "NELEMENTS_TRIANG5[ \t]+"))
      (rx-quad    (make-regexp "NELEMENTS_QUAD1[ \t]+"))
      (rx-l1-node (make-regexp "NODES_LINK1[ \t]*"))
      (rx-l2-node (make-regexp "NODES_LINK2[ \t]*"))
      (rx-l3-node (make-regexp "NODES_LINK3[ \t]*"))
      (rx-l4-node (make-regexp "NODES_LINK4[ \t]*"))
      (rx-l5-node (make-regexp "NODES_LINK5[ \t]*"))
      (rx-t1-node (make-regexp "NODES_TRIANG1[ \t]*"))
      (rx-t3-node (make-regexp "NODES_TRIANG3[ \t]*"))
      (rx-t4-node (make-regexp "NODES_TRIANG4[ \t]*"))
      (rx-t5-node (make-regexp "NODES_TRIANG5[ \t]*"))
      (rx-q-node  (make-regexp "NODES_QUAD1[ \t]*"))
      (rx-list (list (cons rx-point   #f)
                     (cons rx-link1   #t)
                     (cons rx-link2   #t)
                     (cons rx-link3   #t)
                     (cons rx-link4   #t)
                     (cons rx-link5   #t)
                     (cons rx-triang1 #t)
                     (cons rx-triang3 #t)
                     (cons rx-triang4 #t)
                     (cons rx-triang5 #t)
                     (cons rx-quad    #t)
                     (cons rx-coord   #f)
               )
      )
      (m #f)
      (npoint 0)
      (nlink1 0) (nlink2 0) (nlink3 0) (nlink4 0) (nlink5 0)
      (ntriang1 0) (ntriang3 0) (ntriang4 0) (ntriang5 0)
      (nquad 0)
      (node-assoc #f)
    )
    (call-with-current-continuation
      (lambda (return)
        (while (not (null? rx-list))
          (set! m (e-lib-read-keyword fp rx-list))
          (if m
            (let* 
              ( (str (match:string m)) )
              (cond
                ; set number of meshpoints
                ( (regexp-exec rx-point str)
                  (set! npoint (string->number (substring str (match:end m))))
                  (if (not npoint)
                    (begin
                      (display "number of meshpoints is not a number")(newline)
                      (return #f)
                    )
                  )
                  (set! rx-list (cdr (member (cons rx-point #f) rx-list)))
                )
                ; set number of links
                ( (regexp-exec rx-link1 str)
                  (set! nlink1 (string->number (substring str (match:end m))))
                  (if (not nlink1)
                    (begin
                      (display "number of LINK1 elements is not a number")
                      (newline)
                      (return #f)
                    )
                  )
                  (set! rx-list (append (cdr (member (cons rx-link1 #t) 
                                                     rx-list)
                                        )
                                        (list (cons rx-l1-node #f))))
                )
                ( (regexp-exec rx-link2 str)
                  (set! nlink2 (string->number (substring str (match:end m))))
                  (if (not nlink2)
                    (begin
                      (display "number of LINK2 elements is not a number")
                      (newline)
                      (return #f)
                    )
                  )
                  (set! rx-list (append (cdr (member (cons rx-link2 #t) 
                                                     rx-list)
                                        )
                                        (list (cons rx-l2-node #f))))
                )
                ( (regexp-exec rx-link3 str)
                  (set! nlink3 (string->number (substring str (match:end m))))
                  (if (not nlink3)
                    (begin
                      (display "number of LINK3 elements is not a number")
                      (newline)
                      (return #f)
                    )
                  )
                  (set! rx-list (append (cdr (member (cons rx-link3 #t) 
                                                     rx-list)
                                        )
                                        (list (cons rx-l3-node #f))))
                )
                ( (regexp-exec rx-link4 str)
                  (set! nlink4 (string->number (substring str (match:end m))))
                  (if (not nlink4)
                    (begin
                      (display "number of LINK4 elements is not a number")
                      (newline)
                      (return #f)
                    )
                  )
                  (set! rx-list (append (cdr (member (cons rx-link4 #t) 
                                                     rx-list)
                                        )
                                        (list (cons rx-l4-node #f))))
                )
                ( (regexp-exec rx-link5 str)
                  (set! nlink5 (string->number (substring str (match:end m))))
                  (if (not nlink5)
                    (begin
                      (display "number of LINK5 elements is not a number")
                      (newline)
                      (return #f)
                    )
                  )
                  (set! rx-list (append (cdr (member (cons rx-link5 #t) 
                                                     rx-list)
                                        )
                                        (list (cons rx-l5-node #f))))
                )
                ; set number of triangs
                ( (regexp-exec rx-triang1 str)
                  (set! ntriang1 (string->number (substring str (match:end m))))
                  (if (not ntriang1)
                    (begin
                      (display "number of TRIANG1 elements is not a number")
                      (newline)
                      (return #f)
                    )
                  )
                  (set! rx-list (append (cdr (member (cons rx-triang1 #t) 
                                                     rx-list)
                                        )
                                        (list (cons rx-t1-node #f))))
                )
                ( (regexp-exec rx-triang3 str)
                  (set! ntriang3 (string->number (substring str (match:end m))))
                  (if (not ntriang3)
                    (begin
                      (display "number of TRIANG3 elements is not a number")
                      (newline)
                      (return #f)
                    )
                  )
                  (set! rx-list (append (cdr (member (cons rx-triang3 #t) 
                                                     rx-list)
                                        )
                                        (list (cons rx-t3-node #f))))
                )
                ( (regexp-exec rx-triang4 str)
                  (set! ntriang4 (string->number (substring str (match:end m))))
                  (if (not ntriang4)
                    (begin
                      (display "number of TRIANG4 elements is not a number")
                      (newline)
                      (return #f)
                    )
                  )
                  (set! rx-list (append (cdr (member (cons rx-triang4 #t) 
                                                     rx-list)
                                        )
                                        (list (cons rx-t4-node #f))))
                )
                ( (regexp-exec rx-triang5 str)
                  (set! ntriang5 (string->number (substring str (match:end m))))
                  (if (not ntriang5)
                    (begin
                      (display "number of TRIANG5 elements is not a number")
                      (newline)
                      (return #f)
                    )
                  )
                  (set! rx-list (append (cdr (member (cons rx-triang5 #t) 
                                                     rx-list)
                                        )
                                        (list (cons rx-t5-node #f))))
                )
                ; set number of quads
                ( (regexp-exec rx-quad str)
                  (set! nquad (string->number (substring str (match:end m))))
                  (if (not nquad)
                    (begin
                      (display "number of quads is not a number")(newline)
                      (return #f)
                    )
                  )
                  (set! rx-list (append (cdr (member (cons rx-quad #t) 
                                                     rx-list)
                                        )
                                        (list (cons rx-q-node #f))))
                )
                ; read the coordinates
                ( (regexp-exec rx-coord str)
                  (set! node-assoc (e-lib-read-meshpoints npoint fp))
                  (if (not node-assoc)
                    (return #f)
                  )
                  (set! rx-list (cdr (member (cons rx-coord #f) rx-list)))
                )
                ; read link elements
                ( (regexp-exec rx-l1-node str)
                  (if (not (e-lib-read-links node-assoc nlink1 fp 1))
                    (return #f)
                  )
                  (set! rx-list (cdr (member (cons rx-l1-node #f) rx-list)))
                )
                ( (regexp-exec rx-l2-node str)
                  (if (not (e-lib-read-links node-assoc nlink2 fp 2))
                    (return #f)
                  )
                  (set! rx-list (cdr (member (cons rx-l2-node #f) rx-list)))
                )
                ( (regexp-exec rx-l3-node str)
                  (if (not (e-lib-read-links node-assoc nlink3 fp 3))
                    (return #f)
                  )
                  (set! rx-list (cdr (member (cons rx-l3-node #f) rx-list)))
                )
                ( (regexp-exec rx-l4-node str)
                  (if (not (e-lib-read-links node-assoc nlink4 fp 4))
                    (return #f)
                  )
                  (set! rx-list (cdr (member (cons rx-l4-node #f) rx-list)))
                )
                ( (regexp-exec rx-l5-node str)
                  (if (not (e-lib-read-links node-assoc nlink5 fp 5))
                    (return #f)
                  )
                  (set! rx-list (cdr (member (cons rx-l5-node #f) rx-list)))
                )
                ; read triang elements
                ( (regexp-exec rx-t1-node str)
                  (if (not (e-lib-read-triangs node-assoc ntriang1 fp 1))
                    (return #f)
                  )
                  (set! rx-list (cdr (member (cons rx-t1-node #f) rx-list)))
                )
                ( (regexp-exec rx-t3-node str)
                  (if (not (e-lib-read-triangs node-assoc ntriang3 fp 3))
                    (return #f)
                  )
                  (set! rx-list (cdr (member (cons rx-t3-node #f) rx-list)))
                )
                ( (regexp-exec rx-t4-node str)
                  (if (not (e-lib-read-triangs node-assoc ntriang4 fp 4))
                    (return #f)
                  )
                  (set! rx-list (cdr (member (cons rx-t4-node #f) rx-list)))
                )
                ( (regexp-exec rx-t5-node str)
                  (if (not (e-lib-read-triangs node-assoc ntriang5 fp 5))
                    (return #f)
                  )
                  (set! rx-list (cdr (member (cons rx-t5-node #f) rx-list)))
                )
                ; read quad elements
                ( (regexp-exec rx-q-node str)
                  (if (not (e-lib-read-quads node-assoc nquad fp))
                    (return #f)
                  )
                  (set! rx-list (cdr (member (cons rx-q-node #f) rx-list)))
                )
                ( else 
                  (gui-entry-add-history
                    (string-append "Keyword "
                                   (substring (match:string m)
                                              (match:start m) 
                                              (match:end m))
                                   " is not implemented\n"))
                  (return #f)
                )
              )
            )
            (begin
              (gui-entry-add-history "*ERROR*: No keyword could be read\n")
              (return #f)
            )
          )
;          (display "rx-list")(display rx-list)(newline)
        )
        #t
      )
    )
  )
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (operation-e-lib-filename)
  (case *SELECT_STATE*
    ( (cancel)
      (gui-entry-add-history "\n*Merging cancelled*")
      ; clear selection
      ; graphical update is not necessary
      (selection-clear)
    )
    ( (string)
      ; store filename
      (set! mdf-file *SELECT_ANSWER*)
      ; decide new or merging
      (selection-setup selection-keyword
                       operation-e-lib-merge
                       "Read mesh is [New/Merged] <Merged>:"
                       *OPTION_EMPTY* 'view 
                       #f "New Merged")
    )
  )
)



(define (operation-e-lib-merge)
  (let* 
      ( (filename (string-append mdf-file ".mdf"))
        (fp (catch #t 
                   (lambda () (open-input-file filename))
                   (lambda args #f)))
      )
    (if fp
      (let*
        ( (clear #f) )
        (case *SELECT_STATE*
          ( (cancel)
            (set! clear #f)
          )
          ( (keyword)
            (cond
              ( (string-ci=? *SELECT_ANSWER* "New")
                (set! clear #t)
              )
              ( (string-ci=? *SELECT_ANSWER* "Merged")
                (set! clear #f)
              )
            )
          )
        )
        (if clear
          ; reset objects
          (objects-clear-all)
        )
        ; reset graphics
        (view-reset-graphics)
            
        ; read in the e-lib mesh file
        (if (not (e-lib-read-mesh fp))
          ; undo the operations when there was an error
          (undo-all-operations)
        )
        (close-input-port fp)

        ; initialise screen selection
        (object-gl-selection-buffer)
        ; center of rotation is the center of the object
        (view-center-points #f)
        ; zoom to extension
        (view-zoom)
      )
      (gui-entry-add-history
        (string-append "\nERROR: File \"" filename 
                       "\" could not be opened for reading!"))
    )
  )

  ; clear keywords
  (selection-clear)
  ; update the screen
  (gui-update-gl-window #t)
)



