(define-module (sx-ext e-lib-write))
(use-modules  (ice-9 regex)
              (sx entry) (sx undo) (sx property) (sx basic) 
              (sx selection-basic) (sx view) (sx render)
              (sx gui-gl) (sx gui-entry) (sx gui-optmenu)
)

(export command-e-lib-write)


; add the commands to the sx system
(entry-add-command-list '("e-lib-write"))



(define (command-e-lib-write)
  (selection-setup selection-string
                   operation-e-lib-write
                   "Enter filename (with path, no extension):"
                   *OPTION_EMPTY* 'view #f #f)
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; private functions

; writes the header file of the e-lib format 
(define (e-lib-write-header fp)
  (let* ( (counter 0) )
    (display "TITLE              \"Generated by SX\"\n" fp)
    (set! counter 0)
    (do ((i 0 (1+ i))) ((= i *POINT_UNIQUE*))
      (if (vector-ref *POINTS* i)
        (set! counter (1+ counter))
      )
    )
    (display "NMESHPOINTS        "fp)
    (display counter fp)(newline fp)
    (display "NNODES             "fp)
    (display counter fp)(newline fp)

    (if (> *LINK_UNIQUE* 0)
      (begin
        (set! counter 0)
        (do ((i 0 (1+ i))) ((= i *LINK_UNIQUE*))
          (if (vector-ref *LINKS* i)
            (set! counter (1+ counter))
          )
        )
        (display "NELEMENTS_LINK1    "fp)
        (display counter fp)(newline fp)
      )
    )
    (if (> (vector-length *TRIANGS*) 0)
      (begin
        (set! counter 0)
        (do ((i 0 (1+ i))) ((= i *TRIANG_UNIQUE*))
          (if (vector-ref *TRIANGS* i)
            (set! counter (1+ counter))
          )
        )
        (display "NELEMENTS_TRIANG1  "fp)
        (display counter fp)(newline fp)
      )
    )
    (if (> (vector-length *QUADS*) 0)
      (begin
        (set! counter 0)
        (do ((i 0 (1+ i))) ((= i *QUAD_UNIQUE*))
          (if (vector-ref *QUADS* i)
            (set! counter (1+ counter))
          )
        )
        (display "NELEMENTS_QUAD1    "fp)
        (display counter fp)(newline fp)
      )
    )
    (newline fp)(newline fp)
  )
)


; writes the meshpoints
(define (e-lib-write-meshpoints fp)
  (display "MESHPOINT_COORDINATES\n" fp)
  (let*
    ( (n 1) (co #f) (point-data #f)
      (point-map '())
    )
    (do ((i 0 (1+ i))) ((= i *POINT_UNIQUE*))
      (set! point-data (vector-ref *POINTS* i))
      (if point-data
        (begin
          (set! co (assoc-ref point-data 'coords))
          (display n fp)(display "  " fp)
          (display (car   co) fp)(display "  " fp)
          (display (cadr  co) fp)(display "  " fp)
          (display (caddr co) fp)(newline fp)
          (set! point-map (cons n point-map))
          (set! n (1+ n))
        )
        (set! point-map (cons #f point-map))
      )
    )
    (newline fp)(newline fp)
    point-map
  )
)


; writes the elements, link, triangles, quads
(define (e-lib-write-elements point-map fp)
  (set! point-map (reverse point-map))
  ; links
  (if (> (vector-length *LINKS*) 0)
    (let* ( (n 1) (nodes #f) (elem #f) )
      (display "NODES_LINK1\n"fp)
      (do ((i 0 (1+ i))) ((= i *LINK_UNIQUE*))
        (set! elem (vector-ref *LINKS* i))
        (if elem
          (begin
            (set! nodes (assoc-ref elem 'nodes))
            (display n fp)(display "  " fp)
            (display (list-ref point-map (car nodes)) fp)(display "  " fp)
            (display (list-ref point-map (cadr nodes)) fp)(newline fp)
            (set! n (1+ n))
          )
        )
      )
      (newline fp)(newline fp)
    )
  )
  ; triangles
  (if (> (vector-length *TRIANGS*) 0)
    (let* ( (n 1) (nodes #f) (elem #f) )
      (display "NODES_TRIANG1\n"fp)
      (do ((i 0 (1+ i))) ((= i *TRIANG_UNIQUE*))
        (set! elem (vector-ref *TRIANGS* i))
        (if elem
          (begin
            (set! nodes (assoc-ref elem 'nodes))
            (display n fp)(display "  " fp)
            (display (list-ref point-map (car nodes)) fp)(display "  " fp)
            (display (list-ref point-map (cadr nodes)) fp)(display "  " fp)
            (display (list-ref point-map (caddr nodes)) fp)(newline fp)
            (set! n (1+ n))
          )
        )
      )
      (newline fp)(newline fp)
    )
  )
  ; quads
  (if (> (vector-length *QUADS*) 0)
    (let* ( (n 1) (nodes #f) (elem #f) )
      (display "NODES_QUAD1\n"fp)
      (do ((i 0 (1+ i))) ((= i *QUAD_UNIQUE*))
        (set! elem (vector-ref *QUADS* i))
        (if elem
          (begin
            (set! nodes (assoc-ref elem 'nodes))
            (display n fp)(display "  " fp)
            (display (list-ref point-map (car nodes)) fp)(display "  " fp)
            (display (list-ref point-map (cadr nodes)) fp)(display "  " fp)
            (display (list-ref point-map (caddr nodes)) fp)(display "  " fp)
            (display (list-ref point-map (cadddr nodes)) fp)(newline fp)
            (set! n (1+ n))
          )
        )
      )
      (newline fp)(newline fp)
    )
  )
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(define (operation-e-lib-write)
  (case *SELECT_STATE*
    ( (cancel)
      (gui-entry-add-history "\n*Saving in e-lib format is cancelled*")
    )
    ( (string)
      (let*
        ( (filename (string-append *SELECT_ANSWER* ".mdf"))
          (fp (catch #t 
                     (lambda () (open-output-file filename))
                     (lambda args #f)))
        )
        (if fp
          (let* ( (point-map #f) )
            ; write the header
            (e-lib-write-header fp)
            ; write mesh points
            (set! point-map (e-lib-write-meshpoints fp))
            ; write elements
            (e-lib-write-elements point-map fp)
            ; close the file
            (close-output-port fp)
          )
          (gui-entry-add-history
            (string-append "\nERROR: File \"" filename 
                           "\" could not be opened for writing!")
          )
        )
      )
    )
  )
  ; clear keywords
  (selection-clear)
)




