(define-module (sx-ext mirror))
(use-modules  (sx entry) (sx selection-basic) (sx selection)
              (sx property) (sx basic) (sx undo)
              (sx gui-gl) (sx gui-entry) (sx gui-optmenu)
)


(export command-mirror)


; add the command to the sx system
(entry-add-command-list '("mirror"))

; three points determining the plane of mirroring
; internal variables !
(define point0 #f)
(define point1 #f)
(define point2 #f)


; command initialisation 
(define (command-mirror)
  (set! point0 #f)
  (set! point1 #f)
  (set! point2 #f)
  (selection-setup selection-keyword
                   operation-mirror-plane
                   "Mirror plane [XZ/YZ/XY/3 points]:"
                   *OPTION_EMPTY* 'view 
                   '(no_empty) "XZ YZ XY 3")
)

; setting up the mirror plane 
(define (operation-mirror-plane)
  (case *SELECT_STATE*
    ( (keyword)
      (cond
        ( (string-ci=? *SELECT_ANSWER* "XZ")
          (set! point0 '(0 0 0))
          (set! point1 '(1 0 0))
          (set! point2 '(0 0 1))
          ; is the mirror plane is displaced ?
          (selection-setup selection-coord
                           operation-mirror-1-point
                           "Point on the mirror plane <0,0,0>:" 
                           *OPTION_POINT*
                           'select_coord
                           #f #f)
        )
        ( (string-ci=? *SELECT_ANSWER* "YZ")
          (set! point0 '(0 0 0))
          (set! point1 '(0 1 0))
          (set! point2 '(0 0 1))
          (selection-setup selection-coord
                           operation-mirror-1-point
                           "Point on the mirror plane <0,0,0>:" 
                           *OPTION_POINT*
                           'select_coord
                           #f #f)
        )
        ( (string-ci=? *SELECT_ANSWER* "XY")
          (set! point0 '(0 0 0))
          (set! point1 '(1 0 0))
          (set! point2 '(0 1 0))
          (selection-setup selection-coord
                           operation-mirror-1-point
                           "Point on the mirror plane <0,0,0>:" 
                           *OPTION_POINT*
                           'select_coord
                           #f #f)
        )
        ; arbitrary three points are required, setup selection 
        ( (string-ci=? *SELECT_ANSWER* "3")
          ; it also clears the selection
          ; setup to select three point objects !
          (selection-setup selection-object
                           operation-mirror-3-point
                           "Select three points of the mirror plane:" 
                           *OPTION_SELECT*
                           'select_obj
                           ; user must select points
                           '(no_empty type_point)
                           #f)
        )
      )
    )
  )
)

; the mirror plane of XZ, YZ, XY can be displaced or
; in other words a point can be given and the plane will go through this point
; empty answer means, the plane goes through the origo
(define (operation-mirror-1-point)
  (case *SELECT_STATE*
    ( (point)
      (let* ( (disp (list *SELECT_X* *SELECT_Y* *SELECT_Z*)) )
        (set! point0 (map (lambda (x y) (+ x y)) point0 disp))
        (set! point1 (map (lambda (x y) (+ x y)) point1 disp))
        (set! point2 (map (lambda (x y) (+ x y)) point2 disp))
      )
    )
  )
  (operation-mirror-object)
)


; store the coordinates of the three selected points 
(define (operation-mirror-3-point)
  (case *SELECT_STATE*
    ( (objects)
      (if (= (length *SELECT_LIST*) 3)
        (let* 
          ( (n1 (car *SELECT_LIST*))
            (n2 (cadr *SELECT_LIST*))
            (n3 (caddr *SELECT_LIST*))
          )
          (set! point0 (object-property-get (car n1) (cadr n1) 'coords))
          (set! point1 (object-property-get (car n2) (cadr n2) 'coords))
          (set! point2 (object-property-get (car n3) (cadr n3) 'coords))
          (operation-mirror-object)
        )
      )
    )
  )
)


; setups the object selection part
(define (operation-mirror-object)
  (selection-clear)
  (gui-update-gl-window #t)
  (selection-setup selection-object
                   operation-mirror-do
                   "Select object(s) to mirror:" 
                   *OPTION_SELECT*
                   'select_obj
                   ; user must select objects
                   '(no_empty type_any)
                   #f)
)


; performs the actual mirroring
(define (operation-mirror-do)
  (case *SELECT_STATE*
    ( (objects)
      (let*
        ; determine the normal of the plane
        ( (v1 (map (lambda (x y) (- x y)) point1 point0))
          (v2 (map (lambda (x y) (- x y)) point2 point0))
          (x1 (car v1))
          (y1 (cadr v1))
          (z1 (caddr v1))
          (x2 (car v2))
          (y2 (cadr v2))
          (z2 (caddr v2))
          (normal1 (list (- (* y1 z2) (* y2 z1))
                         (- (* x2 z1) (* x1 z2))
                         (- (* x1 y2) (* x2 y1))))
          (length (sqrt (+ (* (car normal1) (car normal1))
                           (* (cadr normal1) (cadr normal1))
                           (* (caddr normal1) (caddr normal1)))))
          (normal (list (/ (car normal1) length)
                        (/ (cadr normal1) length)
                        (/ (caddr normal1) length)))
          (ptlist '())
        )
        (undo-add-op-mark)
        ; add referenced points
        (selection-add-ref-nodes #t)
        ; duplicate and mirror the points 
        ; FIXME: this is something like object-copy-internal ?????
        (for-each
          (lambda (x)
            (if (eq? (car x) 'p)
              (let*
                ( (n (cadr x))
                  (co (object-property-get 'p n 'coords))
                  ; vector from point to a point on the plane
                  (v3 (map (lambda (x y) (- x y)) point0 co))
                  ; scalar product of v3 vector and the normal
                  ; it provides the length of v3 projected onto the 
                  ; normal vector
                  (dirlen (* (+ (* (car normal) (car v3))
                                (* (cadr normal) (cadr v3))
                                (* (caddr normal) (caddr v3))) 2.0))
                  ; from the point in the direction of normal with twice
                  ; the distance
                  (xn (+ (car co) (* dirlen (car normal))))
                  (yn (+ (cadr co) (* dirlen (cadr normal))))
                  (zn (+ (caddr co) (* dirlen (caddr normal))))
                )
                ; create new point,
                ; but store old and new point id so elements can be reproduced
                ; correctly
                (set! ptlist (cons (list n 
                                         (point-add-basic
                                          #f 
                                          (list xn yn zn)
                                          (object-property-get-all-user 'p n)
                                          #t)
                                   )
                                   ptlist) )
              )
            )
          )
          *SELECT_LIST*
        )
        ; duplicate the elements
        (for-each
          (lambda (x) (object-duplicate (car x) (cadr x) ptlist) )
          *SELECT_LIST*
        )
      )
    )
  )
  ; clear all selection
  (selection-clear)
  ; update the screen
  (gui-update-gl-window #t)
)




